#include "SettingsForm.h"
#include "..\..\common\fast_bit_operation_macros.h"
#include <string.h>

#pragma warning (disable:4800)

using namespace FilterSpace;
using namespace System::Windows::Forms;



/////////////////////////////
//       FILTER            //
/////////////////////////////

bool FilterCore(filtersettings* f, int sx, int sy, int sw, int sh, Semaphore^ updater)
{
	if      (!f)           return false;
	else if (!(f->datain)) return false;

	// phase = 0 -val kell kezdeni
	byte  nbhood;
	int   w          = f->width;
	int   h          = f->height;
	int   sin        = f->stridein;
	int   sout       = f->strideout;
	int   cnt        = 0;

	//                         deg        0L    0R    90L   90R   180L  180R  270L  270R
	byte morph_rw[8]                  = { 0x07, 0x12, 0x94, 0x50, 0xe0, 0x48, 0x29, 0x0a };
	byte morph_mask_rw[8]             = { 0x18, 0x85, 0x42, 0xa4, 0x18, 0xa1, 0x42, 0x25 }; 

	byte  m          = morph_rw[f->morph_phase];
	byte  mm		 = morph_mask_rw[f->morph_phase];

	sx = INTERVAL_LIMITER(sx, 0, f->width - 1);
	sy = INTERVAL_LIMITER(sy, 0, f->height - 1);
	if (sw < 1) sw = f->width - sx;
	if (sh < 1) sh = f->height - sy;
	sw = INTERVAL_LIMITER(sw, 1, f->width - sx);
	sh = INTERVAL_LIMITER(sh, 1, f->height - sy);
	int processed_pixels = 0;

	byte* img = f->datain + sy * sin;
	byte* p   = f->dataout;

	for (int y = 0, yy = sy; y < sh; y++, yy++)
	{
		memcpy(p, img + (sx >> 3), (sw + 7) >> 3);

		for (int x = 0, xx = sx; x < sw; x++, xx++)
		{
			if (GET_BITPIXEL_OF_ROW(img, xx)) continue;
			GET_NEIGHBOURS_BITPLANE(f->datain, xx, yy, w, h, sin, nbhood);
			if ((((~nbhood ^ m) | mm) & 0xff) != 0xff) continue; // HA NEM ISMERTK FEL A MINTT
			SET_WHITEBITPIXEL_OF_ROW(p,x);
			cnt++;
		}
		img += sin;
		p   += sout;

		// FOR PERFORMANCE MEASUREMENT
		processed_pixels += sw;
		ImageFilterBase::UpdateFilterStatus(f, processed_pixels, updater);
		if (f->nCoreKillSignal) { if (*(f->nCoreKillSignal)) break; }
	}
	f->eliminated_pixels[f->nCoreThreadId] = cnt;

	return true;
}



IntPtr ImageFilter::CreateFilterSettings(Byte* fs, bool silent)
{
	int  stride, width, height, run_local = 1;
	bool cancelled = false;

	if (!fs) return IntPtr(0);

	xml_node ^xt,  ^xn      = gcnew xml_node( fs );
	filtersettings* f       = XmlToMyStruct( xn, true );
	byte*           indata  = f->datain;
	byte*           outdata = f->dataout;

	if (f && !silent)
	{
		FilterTemplate::SettingsForm^ form = gcnew FilterTemplate::SettingsForm(f);
		form->ShowDialog();
		cancelled = form->FilterCancelled;
		delete form;
	}
	if (cancelled)
	{
		delete f;
		delete xn;
		return IntPtr(0);
	} 

	// GET "Run_Local" PARAMETER
	xn->XmlGetChildValueByName("Run_Local", run_local);

	if (!run_local)
	{
		if (outdata == 0 || indata == outdata)  // FIRST RUN OF CreateFilterSettings
		{
			if (f->width * f->height)
			{
				xn->XmlRemoveChildByName("OutputImage0");

				width  = f->width;
				stride = f->stridein;
				height = f->height;
				outdata = new Byte[stride * height];

				xt = gcnew xml_node( xn, "OutputImage0", "", "" );
				xn->children.Add( xt );
				//                     PinName,  width, height, stride, packed, interleaved, channelcount, pixelsize,       data)
				CreateImageSetting(xt, "Output", width, height, stride,      1,           1,            1,         1,    outdata);
				xt->XmlGetChildByName( xt, "ChannelsBitAllocation" );
				SetChannelsBitAllocation(xt, 0, 1);

				byte binary_color_map[] = { 0, 0, 0, 0, 255, 255, 255, 0 };
				String^ colMap = stringToString(base64((unsigned char*)(binary_color_map), 8));
				xt->XmlSetChildValueByName("ColorMap", colMap, false); 
			}
		}
	}
	else
	{
		xn->XmlRemoveChildByName("OutputImage0");
		xn->XmlGetChildByName(xt, "InputImage0");
		Byte* fb = xt->XML_Tree_Marshal();
		xt = gcnew xml_node(fb);
		delete [] fb;
		xt->name = "OutputImage0"; 
		xn->children.Add( xt );
	}

	xn->XmlSetChildValueByName("FilteringIterationsLimit", f->filtering_iterations_limit);

	delete f;
	Byte* res = xn->XML_Tree_Marshal();
	delete xn;
	return (IntPtr)(void*)res;
}



filtersettings* ImageFilter::XmlToMyStruct(xml_node^ xn, bool creating)
{
	int packed, interleaved, iTmp;
	xml_node^ xntmp;

	if (! xn->XmlGetChildByName(xntmp, "InputImage0")) return 0;
	filtersettings* filset = new filtersettings;

	if (GetOneImageSetting(xntmp, filset->width, filset->height, filset->stridein, packed, interleaved, filset->channelcountin, filset->pixelsizein, filset->datain))
	{
		MessageBox::Show("XmlToMyStruct - 0 - " + creating.ToString());
		delete filset;
		return 0;
	}

	if (! xn->XmlGetChildByName(xntmp, "OutputImage0") && !creating)
	{
		MessageBox::Show("XmlToMyStruct - 1 - " + creating.ToString());
		delete filset;
		return 0;
	}

	if (GetOneImageSetting(xntmp, iTmp, iTmp, filset->strideout, packed, interleaved, filset->channelcountout, filset->pixelsizeout, filset->dataout))
	{
		MessageBox::Show("XmlToMyStruct - 2 - " + creating.ToString());
		delete filset;
		return 0;
	}

	int m = 0;
	if ( xn->XmlGetChildByName(xntmp, "Mask") )	xntmp->XmlGetChildValueByName("Data", m);
	filset->mask = (Byte*)(void*)m;

	filset->processing_size            = ( xn->XmlGetChildValueByName("Processing_Size", m) )  ? (int*)(void*)m : 0;
	filset->processing_ptr             = ( xn->XmlGetChildValueByName("Processing_Ptr", m) )   ? (int*)(void*)m : 0;
	filset->nCoreEnabled               = ( xn->XmlGetChildValueByName("nCore!Enabled", m) )    ? m : 0;
	filset->nCoreKillSignal            = ( xn->XmlGetChildValueByName("nCore!KillSignal", m) ) ? (int*)(void*)m : 0;
	filset->nCoreExecutionSize         = ( xn->XmlGetChildValueByName("nCore!ExecutionSize", m) )      ? m : 1;
	filset->filtering_iterations_limit = ( xn->XmlGetChildValueByName("FilteringIterationsLimit", m) ) ? m : 0; // 0 = skeleton

	return filset;
}


bool ImageFilter::CanLocalize(){ return true; }


IntPtr ImageFilter::GetSupportedFormats()
{
	xml_node ^xn, ^node = gcnew xml_node( nullptr, "", "", "" );

	node->children.Add( gcnew xml_node(node, "InputImageCount", "", "1" ) );
	node->children.Add( gcnew xml_node(node, "OutputImageCount", "", "1" ) );
	xn = gcnew xml_node( node, "InputImage0", "", "" );
	node->children.Add( xn );
	//                     PinName, width, height, stride, packed, interleaved, channelcount, pixelsize, data)
	CreateImageSetting(xn, "Input",     0,      0,      0,      1,           1,            1,         1,    0);
	xn->XmlGetChildByName( xn, "ChannelsBitAllocation" );
	SetChannelsBitAllocation(xn, 0, 1);

	xn = gcnew xml_node( node, "OutputImage0", "", "" );
	node->children.Add( xn );
	//                     PinName, width, height, stride, packed, interleaved, channelcount, pixelsize, data)
	CreateImageSetting(xn, "Output",     0,      0,      0,      1,           1,           1,         1,    0);
	xn->XmlGetChildByName( xn, "ChannelsBitAllocation" );
	SetChannelsBitAllocation(xn, 0, 1);
	/*
	width,height = 2 --> width, height must be power of 2... = 8 --> multiply of 8
	packed, interleaved = 0, 1 (true, false)
	channelcount = 0 --> any positive number
	channelcount > 0 --> the value itself
	*/

	return (IntPtr)(void*)node->XML_Tree_Marshal();
}






// ***********************************************************************************************
// DO NOT MODIFY OR MOVE TO ELSEWHERE THESE HELPER FUNCTIONS
// THESE ARE HERE BECAUSE OF FilterCore IMPLEMENTED IN THIS FILE!!!

bool ImageFilter::FilterCore(filtersettings* f, int sx, int sy, int sw, int sh)
{
	return false; // (::FilterCore(f, sx, sy, sw, sh, nullptr));
}


bool ImageFilter::ExecuteFilter(Byte* fs)
{ 
	bool b = false;
	if (!fs ) return false;
	xml_node^ xn = gcnew xml_node( fs );
	filtersettings* filset = XmlToMyStruct( xn, false );
	delete xn;
	if (filset)
	{
		byte *tmp;
		filter_core_func f         = (::FilterCore);
		int  iteration             = 0;
		int  eliminated_pixels     = 1;
		int  cpucores              = System::Environment::ProcessorCount;
		bool parallelized          = (filset->nCoreEnabled > 0) && (cpucores > 1);
		*(filset->processing_size) = filset->width * filset->height;
		filset->nCoreThreadId      = 0;
		filset->nCoreExecutionSize = 1; 
		filset->eliminated_pixels  = new int[cpucores];
		memset(filset->eliminated_pixels, 0, sizeof(int) * cpucores); 

		// A FILTER 1X MINDENKPPEN LEFUT. HA BEFEJEZI A FUTST, AKKOR AZ ELIMINLT PIXELEK SZMT 0-RA LLTJA

		// mr le van foglalva a dataout
		if (filset->datain != filset->dataout) memcpy(filset->dataout, filset->datain, filset->stridein * filset->height);

		// az eredeti kpet nem befolysoljuk
		byte *should_be_erased = filset->datain = new byte[filset->stridein * filset->height];

		// felcserljk a puffereket, a dataout-rl indul s pros szm trssal ott is vgez minden menetben
		tmp				= filset->datain;
		filset->datain	= filset->dataout;
		filset->dataout = tmp;

		b = true;
		while (eliminated_pixels && ((filset->filtering_iterations_limit == 0 && iteration < 40) || (iteration < filset->filtering_iterations_limit)))
		{
			*(filset->processing_ptr) = 0;

			eliminated_pixels = 0; // VOLT MIT CSINLNI?
			for (int rotation = 0; rotation < 8; rotation++)
			{
				filset->morph_phase = rotation;
				b &= (parallelized) ? ImageFilterBase::ExecuteFilterBox(filset, f) : f(filset, 0, 0, 0, 0, nullptr);

				for (int i = 0; i < cpucores; i++) eliminated_pixels += filset->eliminated_pixels[i];

				tmp				= filset->datain;
				filset->datain	= filset->dataout;
				filset->dataout = tmp;
			}
			if (!b) break;
			iteration++;
		}

		delete [] should_be_erased;
		delete [] filset->eliminated_pixels;
		*(filset->processing_ptr) = *(filset->processing_size);
		*(filset->nCoreKillSignal) = 0; // RESET SIGNAL BIT IN THE MAIN PROGRAM
	}
	return b;
}





